/* private_dirty.p
 * 
 * sudo bin/dirtymem.sh
 *  -or-
 * sudo ${DLC}/bin/_progres -b -p util/private_dirty.p
 * 
 * If not run as root you will only get data for processes that you own.
 * 
 * Report on non-shared resident memory per process - this is the unique
 * memory burden of that process. This is the memory that would be returned
 * to the system if the process were to be killed.
 * 
 * We want to keep the sum of this memory plus the sum of all shared memory
 * low enough to avoid any excuse for the OOM killer to act. As a rule of
 * thumb aim to leave 25%-50% of memory free for other purposes.
 * 
 * This report can also be quite helpful when looking for memory leaks over
 * time. If you have processes with slow but persistent memory growth you can
 * see that by comparing hour by hour or day by day samples.
 * 
 * This data is a PITA to collect. On large systems it can take several minutes
 * to parse /proc/<PID>/smaps (surprisingly the 4gl is not much slower than C
 * code to do this, the overhead is probably the OS populating smaps).
 *
 * A very similar, and quite helpful, tool is "smem". "smem" is a Python
 * script. The download is just 60k. It has a lot of very useful options.
 * 
 * To download smem:
 * 
 *    $ sudo apt-get install smem
 *    
 *     - or -
 *    
 *    $ sudo su -
 *    # cd /tmp
 *    # wget http://www.selenic.com/smem/download/smem-1.4.tar.gz
 *    # tar xvzf smem-1.4.tar.gz
 *    # cp /tmp/smem-1.4/smem /usr/local/bin/
 *    # chmod +x /usr/local/bin/smem
 * 
 * A very useful article that explains more:
 *
 * https://gist.github.com/sameo/d49c50772d616ae00e96c9967e676976
 *
 *   From http://unix.stackexchange.com/questions/33381/getting-information-about-a-process-memory-usage-from-proc-pid-smaps
 *   
 *   For a given process in /proc/<pid>/smaps, for a given mapping entry what are:
 *   
 *       Shared_Clean
 *       Shared_Dirty
 *       Private_Clean
 *       Private_Dirty
 *   
 *   Clean pages are pages that have not been modified since they were mapped
 *   (typically, text sections from shared libraries are only read from disk
 *   (when necessary), never modified, so they'll be in shared, clean pages).
 *   
 *   Dirty pages are pages that are not clean (i.e. have been modified).
 *   
 *   Private pages are available only to that process, shared pages are mapped
 *   by other processes*.
 *   
 *   RSS is the total number of pages, shared or not, currently mapped into the
 *   process. So Shared_Clean + Shared_Dirty would be the shared part of the RSS
 *   (i.e. the part of RSS that is also mapped into other processes), and
 *   Private_Clean + Private_Dirty the private part of RSS (i.e. only mapped in
 *   this process).
 *  
 *   *Note that a "share-able" page is counted as a private mapping until it is
 *   actually shared. i.e. if there is only one process currently using libfoo,
 *   that library's text section will appear in the process's private mappings.
 *   It will be accounted in the shared mappings (and removed from the private
 *   ones) only if/when another process starts using that library.
 *  
 */

define temp-table tt_procList no-undo
  field dirty   as decimal   format ">,>>>,>>>,>>9.9999"
  field pid     as character format "x(10)"
  field usrName as character format "x(10)"
  field start   as character format "x(10)"
  field cpuTime as character format "x(10)"
  field cmd     as character format "x(60)"
  index dirty-idx is primary dirty descending
  index pid-idx   is unique  pid
.

define new global shared variable pt_rptdir as character no-undo.

define variable pidFile    as character no-undo format "x(30)".
define variable cmdLine    as character no-undo format "x(60)".

define variable inputFld   as character no-undo extent 64.
define variable inputLine  as character no-undo.

define variable m1         as decimal   no-undo.

define variable dirty      as decimal   no-undo.
define variable pidDirty   as decimal   no-undo.
define variable totalDirty as decimal   no-undo.

define variable lineNum    as integer no-undo.

define variable n as integer no-undo.
define variable i as integer no-undo.
define variable j as integer no-undo.
define variable k as integer no-undo.

define stream procDir.
define stream pidSMAP.
define stream pidCMDLine.
define stream ps.

function rightAlign returns character ( s as character, w as integer ):
  return fill( " ", w - length( s )) + s.
end.


/* setup
 */

pause 0 before-hide.

if pt_rptdir = "" or pt_rptdir = ? then
  do:
    file-info:file-name = "rpt".
    if file-info:full-pathname <> ? and index( file-info:file-type, "d" ) > 0 then
      pt_rptdir = file-info:full-pathname.
     else
      pt_rptdir = ".".
  end.

m1 = 1024.			/* convert KB to MB	*/


/* cycle through the list of active processes
 */

input stream procDir from os-dir( "/proc" ).

PID_Loop: repeat:

  import stream procDir ^ pidFile.

  if pidFile begins "/proc/." or pidFile >= "/proc/:" then next.

  i = i + 1.
  if i modulo 10 = 0 then
    put screen row screen-lines column 1 string( i, ">,>>>,>>>,>>9" ) + " " + string( totalDirty, ">,>>>,>>>,>>9.9999" ).

  /*
   *
   * 00400000-00ebd000 r-xp 00000000 08:01 30284939                           /opt/oe11710/bin/_progres
   * Size:              10996 kB
   * Rss:                4876 kB
   * Pss:                  33 kB
   * Shared_Clean:       4876 kB
   * Shared_Dirty:          0 kB
   * Private_Clean:         0 kB
   * Private_Dirty:         0 kB
   * Referenced:         4876 kB
   * Anonymous:             0 kB
   * AnonHugePages:         0 kB
   * Swap:                  0 kB
   * KernelPageSize:        4 kB
   * MMUPageSize:           4 kB
   * Locked:                0 kB
   * ProtectionKey:         0
   * VmFlags: rd ex mr mp me dw sd
   * ...
   *
   */

  /* add up all of the private dirty pages for a process
   */

  pidDirty = 0.

  /* processes can disappear before we get a chance to open the "smaps" file
   */

  lineNum = 0.

  do on error undo, next PID_Loop:
    input stream pidSMAP from value( pidFile + "/smaps" ).
  end.

  SMAP_Loop: repeat:

    assign
      inputFld  = ?
      inputLine = ?
      dirty     = 0
    .

    do on error undo, leave SMAP_Loop
       on endkey undo, leave SMAP_Loop:

      import stream pidSMAP inputFld.

    end.

    lineNum = lineNum + 1.

    if inputFld[1] = "Private_Dirty:" then
      assign
        dirty      = decimal( inputFld[2] ) / m1
        pidDirty   = pidDirty + dirty
      .

  end.

  input stream pidSMAP close.

  /* create a tt_procList for every process that we could get data for
   */

  if lineNum = 0 then			/* if we couldn't read anything don't create a procList	*/
    next PID_Loop.
   else
    do:
      create tt_procList.
      assign
        tt_procList.dirty = pidDirty
        tt_procList.pid   = entry( 3, pidFile, "/" )
      .
    end.

  totalDirty = totalDirty + pidDirty.

  /* get the command line for this process
   */

  cmdLine = "".

  input stream pidCMDLine from value( pidFile + "/cmdline" ).

  repeat:

    readkey stream pidCMDLine.

    if lastkey < 0 then
      leave.
    if lastkey = 0 then
      cmdLine = cmdLine + " ".
     else
      cmdLine = cmdLine + chr( lastkey ).

  end.

  input stream pidCMDLine close.

  tt_procList.cmd   = cmdLine.

end.

input stream procDir close.


/* get some details from "ps -ef"
 */

input stream ps through value ( "ps -ef" ).

repeat:

  assign
    inputFld  = ?
    inputLine = ?
  .

  /*
   * 
   * wssdba@traxnode4.intcomex.com> ps -ef | more
   * xcrmsole   4145   4083  0 11:19 pts/1717 00:00:15 /usr/dlc/bin/_progres /db/trax//xus61t2 -s 128 -D 400 -B 200 -c 20 -ld sentai -yy 1970 -v6colon -inp 32000 -d mdy -rand 2 -U xcrmsole -P          -p wb/workbook.p -clientlog /home/xcrmsole/ts.log -debugalert
   * xgtjsama   4148   4138  0 14:32 pts/192  00:00:00 /bin/sh /opt/apero/scripts/scrs/wbsh/wbwrapper.sh
   * xusschar   4166   4035  0 Jul16 ?        00:00:00 sshd: xusschar@notty
   * wssdba    54703      1  0 10:49 ?        00:00:49 /usr/dlc/bin/_progres -b -p util/pt3agent.p -param tpl /db/tpl/tpl -pf etc/protop.pf -T /home/wssdba/pt3/tmp -errorstack -clientlog /home/wssdba/pt3/tmp/pt3agent.tpl.debug
   * wssdba    54759      1  6 10:49 ?        00:21:14 /usr/dlc/bin/_progres -b -p util/pt3agent.p -param xus61t2 /db/trax/xus61t2 -pf etc/protop.pf -T /home/wssdba/pt3/tmp -errorstack -clientlog /home/wssdba/pt3/tmp/pt3agent.xus61t2.debug
   * wssdba   163178  57570  0 16:19 ?        00:00:00 /bin/sh -c ps -ef | grep _progres | grep "p util/pt3agent" | grep -v grep | grep "param aztest "
   * wssdba   163181 163178  0 16:19 ?        00:00:00 grep p util/pt3agent
   * 
   */

  do on error undo, next
     on endkey undo, leave:

    /* import stream ps inputFld no-error.	*/

    import stream ps
      inputFld[1]
      inputFld[2]
      inputFld[3]
      inputFld[4]
      inputFld[5]
      inputFld[6]
      inputFld[7]
    no-error.

  end.

  if inputFld[1] = ? then leave.

  /* only update processes that we are successfully tracking
   */

  find tt_procList where tt_procList.pid = trim( inputFld[2] ) no-error.
  if available tt_procList then
    do:

      assign
        tt_procList.usrName = inputFld[1]
        tt_procList.start   = inputFld[5]
        tt_procList.cpuTime = inputFld[7]
        /* tt_procList.cmd     = cmdLine	*/
      .

/*
 *     do j = 8 to 64:
 *       if inputFld[j] = ? then leave.
 *       tt_procList.cmd = tt_procList.cmd + " " + inputFld[j].
 *     end.
 */

       /* tt_procList.cmd = inputLine.		*/

    end.

end.

input stream ps close.


/* report output
 *
 *      9,572 processes,     119.51 MB of private_dirty pages
 * private dirty (MB)  pid         username    started       cpu time  command line
 * ------------------  ----------  ----------  ----------  ----------  ------------------------------------
 *            71.2680        3648  root            2020    7-01:32:48   ./LSA start
 *             8.1115        3804  gdm             2020    5-13:19:07   /usr/bin/gnome-shell
 */

session:date-format = "ymd".

output to value( substitute( "&1/private_dirty.rpt.&2", pt_rptdir, replace( replace( replace( substring( string( now ), 1, 19 ), "/", "." ), " ", "-" ), ":", "." ))).

os-command value( "free -h -w" ).
put skip(1).

os-command value( "cat /proc/meminfo" ).
put skip(1).

put unformatted
  i substitute( " processes, &1MB of private_dirty pages", trim( string( totalDirty, ">,>>>,>>>,>>9.9999" ))) skip
.

put skip(1).

put unformatted
  "private dirty (MB)       pid    username     started      cpu time  command line"
  skip
  "------------------  ----------  ----------  ----------  ----------  ------------------------------------"
  skip
.

for each tt_procList:

  put
    tt_procList.dirty
    space(2)
    rightAlign( tt_procList.pid, 8 ) format "x(10)"
    space(2)
    tt_procList.usrName
    space(2)
    rightAlign( tt_procList.start, 8 ) format "x(10)"
    space(2)
    rightAlign( tt_procList.cpuTime, 10 ) format "x(10)"
    space(2)
  .

  put unformatted
    tt_procList.cmd
    skip
  .

end.

output close.

quit.
